# Abaqus read plugin .inp mesh

namespace eval Abaqus {
    variable info_plugin ;#information read from .xml plugin file
    #internal use variables
    variable percent_done
    variable stop
    variable _merge  1 ;#to set in file browser options if want to merge or delete old model+results
    variable debug 0

    set info_plugin(path) [file dirname [info script]] ;#to know where to find the files
    if { [info exists info_plugin(Version)] } {
        set other_plugin_version $info_plugin(Version)
    }
    array set info_plugin [ReadProblemtypeXml [file join $info_plugin(path) [file tail $info_plugin(path)].xml] InfoPlugin {Name Version MinimumGiDVersion}]
    if { [info exists other_plugin_version] && $other_plugin_version != $info_plugin(Version)} {
        WarnWinText [_ "There are two versions of %s: the %s and %s inside the plugins folder. One must be deleted" $info_plugin(Name) $other_plugin_version $info_plugin(Version)]
    }
    if { [GidUtils::VersionCmp $info_plugin(MinimumGiDVersion)] < 0 } {
        WarnWinText [_ "This plugin requires GiD %s or later" $info_plugin(MinimumGiDVersion)]
    }
}

#to add more options to the standar file browser
proc Abaqus::MoreImportOptions { f } {
    variable _merge
    ttk::checkbutton $f.chkGeometry -text [_ "Merge"] -variable Abaqus::_merge
    grid columnconfigure $f 1 -weight 1
    grid $f.chkGeometry -sticky w -columnspan 2

    variable export_dir
    if { ![info exists export_dir] } {
        #trick to show the more options
        set export_dir open
    }
    return Abaqus::export_dir
}

proc Abaqus::ReadPreW { } {
    variable _merge
    set multiple 1
    set filenames [MessageBoxGetFilename file read [_ "Read Abaqus mesh"] \
        {} [list [list {Abaqus} {.inp }] [list [_ "All files"] {.*}]] {} $multiple \
        [list [_ "Import options"] Abaqus::MoreImportOptions]]
    if { [llength $filenames] } {
        if { !$_merge } {
            GiD_Process Mescape Files New No escape ;#to avoid previous geometry
        }
        GidUtils::EvalAndEnterInBatch Abaqus::ReadPre $filenames
    }
}


#to drag and drop a file
proc Abaqus::ReadPreSingleFile { filename } {
    variable _merge
    if { !$_merge } {
        GiD_Process Mescape Files New No escape ;#to avoid previous geometry
    }
    GidUtils::EvalAndEnterInBatch Abaqus::ReadPre [list $filename]
}

proc Abaqus::ReadPre { filenames } {
    set fail 0
    GidUtils::WaitState .gid
    GidUtils::DisableGraphics
    if { [catch { set fail [Abaqus::ReadPre0 $filenames] } err options] } {
        set fail 1
    }
    GidUtils::EnableGraphics
    GidUtils::EndWaitState .gid
    if { $fail } {
        return -options $options $err
    } else {
        MeshView 0
        GiD_Process 'Zoom Frame escape
        GidUtils::UpdateWindow LAYER
        GidUtils::UpdateWindow GROUPS
        GidUtils::SetWarnLine [_ "Geometry read"]
    }
    return $fail
}

proc Abaqus::GetGiDElementTypeAndNnodes { abaqus_element_type } {
    variable debug
    set linear_2 {B21 B21H B23 B23H B31 B31H GK3D2 GK3D2N PIPE21 PIPE21H SAX1 T2D2 T2D2H T2D3 T2D3H T3D2 T3D3H}
    set linear_3 {B22 B22H B32 B32H B32R B33 B33H B34 D PIPE31 PIPE31H}
    set triangles_3 {CPS3 DS3 M3D3 R3D3 S3 S3R STRI3 STRI35}
    set triangles_6 {CAX6 CPE6 DS6 M3D6 S6 STRI65}
    set quadrilateral_4 {CAX4P CPS4 CPS4R DS4 M3D4 M3D4R R3D4 S4 S4R S4RS S4RSW S4R5}
    set quadrilateral_8 {CAX8 CAX8R CPS8 CPS8R DS8 M3D8 M3D8R S8 S8R S8R5 S8RT}
    set quadrilateral_9 {S9 S9R5}
    set tetrahedra_4 {C3D4 C3D4E C3D4H DC3D4 DC3D4E}
    set tetrahedra_10 {C3D10 C3D10E C3D10H C3D10I C3D10M C3D10MH DC3D10 DC3D10E}
    set hexahedra_8 {AC3D8 C3D8 C3D8E C3D8H C3D8HT C3D8I C3D8IH C3D8R C3D8RH C3D8T DC3D8 DC3D8E DCC3D8D DK3D8N F3D8 GK3D8}
    set hexahedra_20 {AC3D20 C3D20 C3D20E C3D20H C3D20HT C3D20R C3D20RH C3D20RHT C3D20RT C3D20T CDD20RE DC3D20 DC3D20E}
    set hexahedra_27 {C3D27}
    set prism_6 {C3D6 C3D6E C3D6H DC3D6 DC3D6E F3D6 GK3D6 GK3D6N}
    set prism_12 {GK3D12M}
    set prism_15 {C3D15 C3D15E C3D15H DC3D15 DC3D15E}
    if { [lsearch $linear_2 $abaqus_element_type] != -1  } {
        set elem_type Linear
        set element_nnodes 2
    } elseif { [lsearch $linear_3 $abaqus_element_type] != -1  } {
        set elem_type Linear
        set element_nnodes 3
    } elseif { [lsearch $triangles_3 $abaqus_element_type] != -1  } {
        set elem_type Triangle
        set element_nnodes 3
    } elseif { [lsearch $triangles_6 $abaqus_element_type] != -1  } {
        set elem_type Triangle
        set element_nnodes 6
    } elseif { [lsearch $quadrilateral_4 $abaqus_element_type] != -1  } {
        set elem_type Quadrilateral
        set element_nnodes 4
    } elseif { [lsearch $quadrilateral_8 $abaqus_element_type] != -1  } {
        set elem_type Quadrilateral
        set element_nnodes 8
    } elseif { [lsearch $quadrilateral_9 $abaqus_element_type] != -1  } {
        set elem_type Quadrilateral
        set element_nnodes 9
    } elseif { [lsearch $tetrahedra_4 $abaqus_element_type] != -1  } {
        set elem_type Tetrahedra
        set element_nnodes 4
    } elseif { [lsearch $tetrahedra_10 $abaqus_element_type] != -1  } {
        set elem_type Tetrahedra
        set element_nnodes 10
    } elseif { [lsearch $hexahedra_8 $abaqus_element_type] != -1  } {
        set elem_type Hexahedra
        set element_nnodes 8
    } elseif { [lsearch $hexahedra_20 $abaqus_element_type] != -1  } {
        set elem_type Hexahedra
        set element_nnodes 20
    } elseif { [lsearch $hexahedra_27 $abaqus_element_type] != -1  } {
        set elem_type Hexahedra
        set element_nnodes 27
    } elseif { [lsearch $prism_6 $abaqus_element_type] != -1  } {
        set elem_type Prism
        set element_nnodes 6
    } elseif { [lsearch $prism_12 $abaqus_element_type] != -1  } {
        set elem_type Prism
        set element_nnodes 12
    } elseif { [lsearch $prism_15 $abaqus_element_type] != -1  } {
        set elem_type Prism
        set element_nnodes 15
    } else {
        W "Abaqus::GetGiDElementTypeAndNnodes. Unexpected element type $abaqus_element_type"
        set elem_type ""
        set element_nnodes 0
    }
    return [list $elem_type $element_nnodes]
}

proc Abaqus::CloneSubgroupInGroup { group_name sub_group_name } {
    if { [GiD_Groups exists $sub_group_name] } {
        set tmp_group [Groups::GetAutomaticName ""]
        GiD_Groups clone $sub_group_name $tmp_group
        GiD_Groups edit parent $tmp_group $group_name
        set full_tmp_group [Groups::NameJoin $group_name $tmp_group]
        set full_sub_group_name [Groups::NameJoin $group_name $sub_group_name]
        GiD_Groups edit name $full_tmp_group $full_sub_group_name
    }
    return 0
}

proc Abaqus::GetListFromLine { line } {
    set list_items [list]
    foreach item [split $line ,] {
        lappend list_items [string trim $item]
    }
    return $list_items    
}

proc Abaqus::ReadPreMesh { filename } {
    variable debug
    set fail 0
    set keys_to_ignore [lsort -dictionary {AMPLITUDE ASSEMBLY {BEAM SECTION} BOUNDARY BOUNDARYF BUCKLE CFLUX
    {CHANGE FRICTION} {CHANGE MATERIAL} {CHANGE PLASTIC} CLOAD CONDUCTIVITY
    {CONTACT FILE} {CONTACT PAIR} CONTROLS {COUPLED TEMPERATURE-DISPLACEMENT} CREEP {CYCLIC HARDENING}
    DAMPING {DEFORMATION PLASTICITY} DENSITY DEPVAR DFLUX DLOAD DYNAMIC
    {EL FILE} {EL PRINT} ELASTIC ELSET {END ASSEMBLY} {END INSTANCE} {END STEP} ENDSTEP EQUATION EXPANSION {FACE PRINT} FILM {FLUID CONSTANTS} {FLUID SECTION}
    FREQUENCY FRICTION HEADING {HEAT TRANSFER} HYPERELASTIC HYPERFOAM
    {INITIAL CONDITIONS} INSTANCE MATERIAL MEMBRANE {MODAL DAMPING} {MODAL DYNAMIC} {MODEL CHANGE} MPC {NO ANALYSIS}
    {NODE FILE} {NODE PRINT} ORIENTATION {PHYSICAL CONSTANTS} PLASTIC RADIATE RESTART {RIGID BODY}
    {SHELL SECTION} {SOLID SECTION} {SPECIFIC GAS CONSTANT} {SPECIFIC HEAT} STATIC {STEADY STATE DYNAMICS} STEP
    SURFACE {SURFACE BEHAVIOR} {SURFACE INTERACTION} TEMPERATURE
    TIE {TIME POINTS} TRANSFORM TRANSFORMF {USER MATERIAL} {VALUES AT INFINITY} VIEWFACTOR VISCO}]
    #offset_nodes to be added to the number if a previous model exists
    set offset_nodes [GiD_Info Mesh MaxNumNodes]
    set offset_elements [GiD_Info Mesh MaxNumElements]
    set layers_created [list]
    set layer [file rootname [file tail $filename]]
    if { ![GiD_Layers exists $layer] } {
        GiD_Layers create $layer
        lappend layers_created $layer
    }
    GiD_Layers edit to_use $layer

    set fp [open $filename r]
    if { $fp == "" } {
        WarnWin [_ "Cannot open file '%s'" $filename]
        return 1
    }
    set must_read 1
    while { ![eof $fp] } {
        if { $must_read } {
            gets $fp line
        } else {
            set must_read 1
        }
        if { $line == "" } {
            continue
        }
        if { [string range $line 0 1] == "**" } {
            #comment
            continue
        }
        if { [string index $line 0] == "*" } {
            set items [split $line ,]
            set key [string toupper [string trim [string range [lindex $items 0] 1 end]]]
            #extra attributes
            array unset abaqus_key_attributes
            foreach item [lrange $items 1 end] {
                lassign [split $item =] subkey value
                set abaqus_key_attributes([string toupper [string trim $subkey]]) [string trim $value]
            }
            if { $key == "NODE" } {
                #attributes: nset
                set group_name ""
                if { [info exists abaqus_key_attributes(NSET)] } {
                    set group_name $abaqus_key_attributes(NSET)
                }
                set group_items [list]
                while { ![eof $fp] } {
                    gets $fp line
                    if { $line == "" } {
                        continue
                    }
                    if { [string index $line 0] == "*" } {
                        if { [string range $line 0 1] == "**" } {
                            #comment
                            continue
                        } else {
                            if { $group_name != "" && [llength $group_items] } {
                                if { ![GiD_Groups exists $group_name] } {
                                    GiD_Groups create $group_name
                                }
                                GiD_EntitiesGroups assign $group_name nodes $group_items
                            }
                            set must_read 0
                            break
                        }
                    } else {
                        lassign [string map {, { }} $line] node_id x y z
                        if { $z == "" } {
                            set z 0
                        }
                        if { $offset_nodes } {
                            incr node_id $offset_nodes
                        }
                        lappend group_items [GiD_Mesh create node $node_id [list $x $y $z]]
                    }
                }
            } elseif { $key == "ELEMENT" } {
                #attributes: type elset
                lassign [Abaqus::GetGiDElementTypeAndNnodes [string toupper $abaqus_key_attributes(TYPE)]] elem_type element_nnodes
                set group_name ""
                if { [info exists abaqus_key_attributes(ELSET)] } {
                    set group_name $abaqus_key_attributes(ELSET)
                }
                set group_items [list]
                while { ![eof $fp] } {
                    gets $fp line
                    if { $line == "" } {
                        continue
                    }
                    if { [string index $line 0] == "*" } {
                        if { [string range $line 0 1] == "**" } {
                            #comment
                            continue
                        } else {
                            #GiD_MeshPre_Create $elem_type $element_nnodes $node_ids $vertices $element_ids $element_vertex_indices $zero_based_array $layer
                            if { $group_name != "" && [llength $group_items] } {
                                if { ![GiD_Groups exists $group_name] } {
                                    GiD_Groups create $group_name
                                }
                                GiD_EntitiesGroups assign $group_name elements $group_items
                            }
                            set must_read 0
                            break
                        }
                    } else {
                        while { [string index $line end] == "," && ![eof $fp] } {
                            gets $fp line_extra
                            set line $line$line_extra
                        }
                        set element_items [string map {, { }} $line]
                        set element_id [lindex $element_items 0]
                        set element_connectivities [list]
                        foreach node_id [lrange $element_items 1 end] {
                            if { $offset_nodes } {
                                incr node_id $offset_nodes
                            }
                            lappend element_connectivities $node_id
                        }
                        if { $elem_type != "" } {
                            if { $elem_type == "Hexahedra" && $element_nnodes == 20 } {
                                #Abaqus has different connectivities that GiD
                                set block_1 [lrange $element_connectivities 0 11]
                                set block_2 [lrange $element_connectivities 12 15]
                                set block_3 [lrange $element_connectivities 16 19]
                                set element_connectivities [concat $block_1 $block_3 $block_2]
                            } elseif { $elem_type == "Prism" && $element_nnodes == 15 } {
                                #Abaqus has different connectivities that GiD
                                set block_1 [lrange $element_connectivities 0 8]
                                set block_2 [lrange $element_connectivities 9 11]
                                set block_3 [lrange $element_connectivities 12 14]
                                set element_connectivities [concat $block_1 $block_3 $block_2]
                            }
                            if { $offset_elements } {
                                incr element_id $offset_elements
                            }
                            #instead append use element_id, to have the same id used for elset -> groups
                            if { [catch { lappend group_items [GiD_Mesh create element $element_id $elem_type $element_nnodes $element_connectivities] } err] } {
                                W $err
                            }
                        }
                    }
                }
            } elseif { $key == "NSET" } {
                #attributes: nset
                set group_name ""
                if { [info exists abaqus_key_attributes(NSET)] } {
                    set group_name $abaqus_key_attributes(NSET)
                    if { ![GiD_Groups exists $group_name] } {
                        GiD_Groups create $group_name
                    }
                }
                set group_items [list]
                while { ![eof $fp] } {
                    gets $fp line
                    if { $line == "" } {
                        continue
                    }
                    if { [string index $line 0] == "*" } {
                        if { [string range $line 0 1] == "**" } {
                            #comment
                            continue
                        } else {
                            if { $group_name != "" && [llength $group_items] } {
                                GiD_EntitiesGroups assign $group_name nodes $group_items
                            }
                            set must_read 0
                            break
                        }
                    } else {                        
                        set items [Abaqus::GetListFromLine $line]
                        if { [info exists abaqus_key_attributes(GENERATE)] } {
                            lassign $items from to increment
                            for {set i $from} {$i<=$to} {incr i $increment} {
                                set item $i
                                if { $offset_nodes } {
                                    incr item $offset_nodes
                                }
                                lappend group_items $item
                                
                            }                            
                            
                        } else {                        
                            foreach item $items {
                                if { [string is integer -strict $item] } {
                                    if { $offset_nodes } {
                                        incr item $offset_nodes
                                    }
                                    lappend group_items $item
                                } else {
                                    Abaqus::CloneSubgroupInGroup $group_name $item
                                }       
                            }             
                        }
                    }
                }
            } elseif { $key == "ELSET" } {
                #attributes: elset
                set group_name ""
                if { [info exists abaqus_key_attributes(ELSET)] } {
                    set group_name $abaqus_key_attributes(ELSET)
                    if { ![GiD_Groups exists $group_name] } {
                        GiD_Groups create $group_name
                    }
                }
                set group_items [list]
                while { ![eof $fp] } {
                    gets $fp line
                    if { $line == "" } {
                        continue
                    }
                    if { [string index $line 0] == "*" } {
                        if { [string range $line 0 1] == "**" } {
                            #comment
                            continue
                        } else {
                            if { $group_name != "" && [llength $group_items] } {
                                GiD_EntitiesGroups assign $group_name elements $group_items
                            }
                            set must_read 0
                            break
                        }
                    } else {
                        #expects a list of existing elset (converted to groups)
                        set items [Abaqus::GetListFromLine $line]
                        if { [info exists abaqus_key_attributes(GENERATE)] } {
                            lassign $items from to increment
                            for {set i $from} {$i<=$to} {incr i $increment} {
                                set item $i
                                if { $offset_nodes } {
                                    incr item $offset_nodes
                                }
                                lappend group_items $item
                                
                            }                            
                            
                        } else {                            
                            foreach item $items {
                                if { [string is integer -strict $item] } {                            
                                    if { $offset_elements } {
                                        incr item $offset_elements
                                    }
                                    lappend group_items $item
                                } else {
                                    Abaqus::CloneSubgroupInGroup $group_name $item
                                }
                            }
                        }
                    }
                }
            } elseif { $key == "PART" } {
                set part_name $abaqus_key_attributes(NAME)
                #each part has local numeration, must use offset to not repeat ids
                set offset_nodes [GiD_Info Mesh MaxNumNodes]
                set offset_elements [GiD_Info Mesh MaxNumElements]
                set layer $part_name
                if { ![GiD_Layers exists $layer] } {
                    GiD_Layers create $layer
                    lappend layers_created $layer
                }
                GiD_Layers edit to_use $layer
            } elseif { $key == "END PART" } {
                #can for example assign a group to store information of parts

            } elseif { $key == "INCLUDE" } {
                set include_input $abaqus_key_attributes(INPUT)
                if { [file pathtype $include_input] == "relative" } {
                    set include_input [file join [file dirname $filename] $include_input]
                }
                Abaqus::ReadPreMesh $include_input
            } elseif { [lsearch -dictionary -sorted $keys_to_ignore $key] != -1 } {
                #attributes:
                while { ![eof $fp] } {
                    gets $fp line
                    if { [string index $line 0] == "*" } {
                        if { [string range $line 0 1] == "**" } {
                            #comment
                            continue
                        } else {
                            set must_read 0
                            break
                        }
                    } else {
                        #do something...
                    }
                }
            } elseif { $key == "*" } {
                #comment
            } elseif { $key == "" } {
                #???
            } else {
                set unimplemented_types($key) 1
            }
        } else {
            W "Abaqus::ReadPre. Ignored line $line"
        }
    }

    close $fp

    #delete layers created if they were unused
    foreach layer $layers_created {
        if { ![GiD_Layers get num_entities $layer] } {
            GiD_Layers delete $layer
        }
    }
    if { [GiD_Layers exists Layer0] && [llength [GiD_Layers list]] > 1 } {
        #delete Layer0 if possible
        if { ![GiD_Layers get num_entities Layer0] && ![GiD_Layers get num_conditions Layer0] } {
            GiD_Layers delete Layer0
        }
    }

    if { $offset_nodes != [GiD_Info Mesh MaxNumNodes] } {
        GiD_ModifiedFileFlag set 1
    }
    if { $debug } {
        if { [llength [array names unimplemented_types]] } {
            W "Unimplemented types: [lsort -dictionary [array names unimplemented_types]]"
        }
    }
    return $fail
}

proc Abaqus::ReadPre0 { filenames } {
    variable percent_done
    variable stop
    set fail 0
    set ::Abaqus::stop 0
    set ::Abaqus::percent_done 0
    set percent_max [llength $filenames]
    ::GidUtils::CreateAdvanceBar [_ "Loading file"] [_ "Process"] ::Abaqus::percent_done $percent_max {set ::Abaqus::stop 1}
    foreach filename $filenames {
        if { $::Abaqus::stop } {
            break
        }
        if { [file exists $filename] } {
            set res ""
            if { [catch { set local_fail [Abaqus::ReadPreMesh $filename] } err options] } {
                #set fail 1
                GidUtils::DeleteAdvanceBar
                return -options $options $err
            } else {
                if { $local_fail } {
                    set fail 1
                }
                GiD_RaiseEvent GiD_Event_AfterChangeMesh [ GiD_Info Mesh NumNodes] [ GiD_Info Mesh NumElements]
                GiD_RaiseEvent GiD_Event_AfterOpenFile $filename ABAQUS_FORMAT $local_fail
            }

        }
        incr ::Abaqus::percent_done
    }
    GidUtils::DeleteAdvanceBar
    return $fail
}

proc Abaqus::AddToMenu { } {
    if { [GidUtils::IsTkDisabled] } {
        return
    }
    if { [GiDMenu::GetOptionIndex Files [list Import "Abaqus..."] PRE] != -1 } {
        return
    }

    #try to insert this menu after the word "Files->Import->XYZ nodes"
    set position [GiDMenu::GetOptionIndex Files [list Import "XYZ nodes"] PRE]
    if { $position == -1 } {
        set position end
    }
    GiDMenu::InsertOption Files [list Import "Abaqus..."] $position PRE Abaqus::ReadPreW "" "" insertafter _
}


#invoke this menus changes
Abaqus::AddToMenu
GiDMenu::UpdateMenus

#register the proc to be automatically called when re-creating all menus (e.g. when doing files new)
GiD_RegisterPluginAddedMenuProc Abaqus::AddToMenu

#register the proc to be automatically called when dropping a file
GiD_RegisterExtensionProc ".Abaqus .inp" PRE Abaqus::ReadPreSingleFile
